
import json
import logging
from dataclasses import dataclass
from datetime import datetime
from abc import ABC, abstractmethod

import aiohttp
import validators
from discord import Embed, Colour
from bs4 import BeautifulSoup as bs4
from feedparser import FeedParserDict, parse
from markdownify import markdownify
from textwrap import shorten
from feedparser import parse

from mutators import registry as mutator_registry
from utils import get_unparsed_feed
from api import API

log = logging.getLogger(__name__)
dumps = lambda _dict: json.dumps(_dict, indent=8)


class RSSItem:
    def __init__(self, title, link, description, pub_date, guid, image_url):
        self.title = title
        self.link = link
        self.description = description
        self.pub_date = pub_date
        self.guid = guid
        self.image_url = image_url

    def __str__(self):
        return self.title

    def create_mutated_copy(self, mutators):
        pass

    async def to_embed(self, sub, feed, session):

        title = shorten(markdownify(self.title, strip=["img", "a"]), 256)
        desc = shorten(markdownify(self.description, strip=["img"]), 4096)

        author = ""
        author_url = ""

        icon_url = ""
        thumb_url = ""

        # Replace HTML with Markdown, and shorten text.
        # author = shorten(self.source.name, 256)

        # validate urls
        # author_url = self.source.url if validators.url(self.source.url) else None
        # icon_url = self.source.icon_url if validators.url(self.source.icon_url) else None

        # Combined length validation
        # Can't exceed combined 6000 characters, [400 Bad Request] if failed.
        combined_length = len(title) + len(desc) + (len(author) * 2)
        cutoff = combined_length - 6000
        desc = shorten(desc, cutoff) if cutoff > 0 else desc

        embed = Embed(
            title=title,
            description=desc,
            # timestamp=self.published,
            url=self.link if validators.url(self.link) else None,
            colour=Colour.from_str("#" + sub.embed_colour)
        )

        log.debug("has image url without search: '%s'", self.image_url)

        if sub.article_fetch_image:
            embed.set_image(url=self.image_url or await self.get_thumbnail_url(session))
            embed.set_thumbnail(url=feed.image_href if validators.url(feed.image_href) else None)

        # embed.set_image(url=thumb_url)
        # embed.set_author(url=author_url, name=author)
        # embed.set_footer(text=self.author)

        return embed

    async def get_thumbnail_url(self, session: aiohttp.ClientSession) -> str | None:
        """Returns the thumbnail URL for an article.

        Parameters
        ----------
        session : aiohttp.ClientSession
            A client session used to get the thumbnail.

        Returns
        -------
        str or None
            The thumbnail URL, or None if not found.
        """

        # log.debug("Fetching thumbnail for article: %s", self)

        try:
            async with session.get(self.link, timeout=15) as response:
                html = await response.text()
        except aiohttp.InvalidURL as error:
            log.error("invalid thumbnail url: %s", error)
            return None

        soup = bs4(html, "html.parser")
        image_element = soup.select_one("meta[property='og:image']")
        if not image_element:
            return None

        image_content = image_element.get("content")
        return image_content if validators.url(image_content) else None

import json
class RSSFeed:
    def __init__(self, title, link, description, language='en-gb', pub_date=None, last_build_date=None, image_href=None):
        self.title = title
        self.link = link
        self.description = description
        self.language = language
        self.pub_date = pub_date
        self.last_build_date = last_build_date
        self.image_href = image_href
        self.items = []

    def add_item(self, item: RSSItem):
        if not isinstance(item, RSSItem):
            raise TypeError("item must be an instance of RSSItem")

        self.items.append(item)

    def __str__(self):
        return self.guid

    @classmethod
    def from_parsed_feed(cls, parsed_feed):
        title = parsed_feed.feed.get('title', 'No title')
        link = parsed_feed.feed.get('link', 'No link')
        description = parsed_feed.feed.get('description', 'No description')
        language = parsed_feed.feed.get('language', 'en-gb')
        pub_date = parsed_feed.feed.get('published', None)
        last_build_date = parsed_feed.feed.get('updated', None)
        image_href = parsed_feed.get("image", {}).get("href")

        feed = cls(title, link, description, language, pub_date, last_build_date, image_href)

        for entry in parsed_feed.entries:
            item_title = entry.get('title', 'No title')
            item_link = entry.get('link', 'No link')
            item_description = entry.get('description', 'No description')
            item_pub_data = entry.get('published_parsed', None)
            item_guid = entry.get('id', None) or entry.get("guid", None)
            item_image_url = entry.get("media_content", [{}])[0].get("url")

            # item_published = datetime(*entry.published_parsed[0:-2]) if item_pub_data else None

            item = RSSItem(item_title, item_link, item_description, item_pub_data, item_guid, item_image_url)
            feed.add_item(item)

        feed.items.reverse()

        return feed


DATETIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"


@dataclass
class DjangoDataModel(ABC):

    @staticmethod
    @abstractmethod
    def parser(item: dict) -> dict:
        """Overwrite this method to parse types."""
        return item

    @classmethod
    def from_list(cls, data: list[dict]) -> list:
        return [cls(**cls.parser(item)) for item in data]

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**cls.parser(data))


@dataclass(slots=True)
class Subscription(DjangoDataModel):

    id: int
    name: str
    url: str
    guild_id: int
    creation_datetime: datetime
    extra_notes: str
    filters: list[int]
    mutators: dict[str, list[dict]]
    article_fetch_image: bool
    embed_colour: str
    active: bool
    channels_count: int

    @staticmethod
    def parser(item: dict) -> dict:
        item["guild_id"] = int(item["guild_id"])
        item["creation_datetime"] = datetime.strptime(item["creation_datetime"], "%Y-%m-%dT%H:%M:%S.%f%z")
        item["mutators"] = {
            "title": item.pop("article_title_mutators"),
            "desc": item.pop("article_desc_mutators")
        }

        return item

    async def get_channels(self, api):
        channel_data, _ = await api.get_subscription_channels(subscription=self.id)
        return SubChannel.from_list(channel_data)


@dataclass(slots=True)
class SubChannel(DjangoDataModel):

    id: int
    channel_id: int
    subscription: int

    @staticmethod
    def parser(item: dict) -> dict:
        item["channel_id"] = int(item["channel_id"])
        item["subscription"] = int(item["subscription"])

        return item

    @property
    def mention(self) -> str:
        return f"<#{self.channel_id}>"


@dataclass(slots=True)
class SubscriptionChannel(DjangoDataModel):

    uuid: str
    id: int
    subscription: Subscription
    creation_datetime: datetime

    @staticmethod
    def parser(item: dict) -> dict:
        
        item["id"] = int(item["id"])
        item["subscription"] = Subscription.from_dict(item.pop("subscription"))
        item["creation_datetime"] = datetime.strptime(item["creation_datetime"], DATETIME_FORMAT)
        return item

    @property
    def mention(self) -> str:
        """
        Returns the `id` as a string in the discord mention format.
        """

        return f"<#{self.id}>"

    @property
    def hyperlinks_string(self) -> str:
        """"""

        api_hyperlink = f"[API]({API.CHANNEL_ENDPOINT}{self.uuid}/)"
        rss_hyperlink = f"[RSS]({self.subscription.rss_url})"
        value = f"{rss_hyperlink} · {api_hyperlink}"

        return value


@dataclass(slots=True)
class TrackedContent(DjangoDataModel):

    uuid: str
    subscription: str
    content_url: str
    creation_datetime: datetime

    @staticmethod
    def parser(item: dict) -> dict:

        item["creation_datetime"] = datetime.strptime(item["creation_datetime"], DATETIME_FORMAT)
        return item
